#include <iostream>
#include <iomanip>
#include <sstream>
#include <cstdlib>

#include <TROOT.h>
#include <TApplication.h>
#include <TFile.h>
#include <TTree.h>

#include "MediumMagboltz.hh"
#include "SolidBox.hh"
#include "GeometrySimple.hh"
#include "ComponentConstant.hh"
#include "Sensor.hh"
#include "TrackHeed.hh"
#include "Plotting.hh"
#include "Random.hh"

#include "Electron.h"
#include "Cluster.h"
#include "Track.h"
#include "MPGD.h"
#include "Hit.h"
#include "Event.h"

using namespace Garfield;

int main(int argc, char * argv[]) {

// Define some constants
  const double GeV  = 1.e9; // GeV to eV conversion factor
  const double rad  = TMath::Pi()/180.; // deg to rad conversion factor

#if 0
  std::cerr << "agrc = " << argc << std::endl;
  for (int i = 0; i < argc; i++) {
    std::cerr << "agrv["  << i << "] = " << argv[i] << std::endl;
  }
#endif

// Read in command line parameters
  int         nEvents  = 100000; // The number of events to generate
  double      dlen     = 25.;    // drift length [cm]
  double      phideg   =  0.;    // phi [deg]
  double      thetadeg =  0.;    // theta [deg]
  double      pGeV     =  5.;    // Momentum [GeV]
  std::string pid("electron");   // Particle name

  switch (argc) {
    case 7:
      pid      = argv[6];
    case 6:
      pGeV     = atof(argv[5]);
    case 5:
      thetadeg = atof(argv[4]);
    case 4:
      phideg   = atof(argv[3]);
    case 3:
      dlen     = atof(argv[2]);
    case 2:
      nEvents  = atoi(argv[1]);
      break;
    default:
      break; 
  }
  std::cerr << "=================================================" << std::endl;
  std::cerr << " nEvents = " << nEvents              << std::endl
            << " dlen    = " << dlen     << " [cm]"  << std::endl
            << " phi     = " << phideg   << " [deg]" << std::endl
            << " theta   = " << thetadeg << " [deg]" << std::endl
            << " p       = " << pGeV     << " [GeV]" << std::endl
            << " pid     = " << pid                  << std::endl;
  std::cerr << "=================================================" << std::endl;

  double      phi     = phideg  *rad;    // phi [rad]
  double      theta   = thetadeg*rad;    // theta [rad]
  double      p       = pGeV    *GeV;    // Momentum [eV]

  randomEngine.Seed(123456); 
  TApplication app("app", &argc, argv); 
  plottingEngine.SetDefaultStyle();

  // Make a medium
  auto gas = new MediumMagboltz();
  gas->SetComposition("ar", 95., "c4h10", 2., "cf4", 3.);
  gas->SetTemperature(293.15);
  gas->SetPressure(760.);

  // Detector geometry
#if 0
  GEN::MPGD tpc; // lp1
#else
  //--
  // Change tpc properties here if needed
  //--
  GEN::MPGD tpc(250.,   // Maximum Drift Length [cm]                      
                0.116,  // Pad Pitch [cm]
                0.536,  // Pad Row Pitch [cm]
                7.5e-3, // Drift Velocity [cm/ns]
                95.e-4, // Transverse Diffusion Constant [cm/sqrt(cm)]
#if 0
                300.e-4,// Longitudinal Diffusion Constant [cm/sqrt(cm)]
#else
                220.e-4,// Longitudinal Diffusion Constant [cm/sqrt(cm)]
#endif
                2000.,  // Gas Gain Average
#if 1
                0.6,    // Gas Gain Fluctuation Parameter (Polya)       
#else
                100.,   // Gas Gain Fluctuation Parameter (Polya)       
#endif
#endif
#if 0
                0.001); // PRF width [cm]
#else
                0.027); // PRF width [cm]
#endif
  
  const int    nrows  = 4; // track length in uinits of pad row pitch
  const double width  = nrows*tpc.GetPadRowPitch(); // track length [cm]
  const double zwidth = tpc.GetMaxDriftLength();    // tpc z length [cm] 
  auto box = new SolidBox(0., 0., 0., 10., width/2., zwidth);
  auto geo = new GeometrySimple();
  geo->AddSolid(box, gas);


  // Make a component
  auto comp = new ComponentConstant();
  comp->SetGeometry(geo);
  comp->SetElectricField(0., 0., 230.); // E [kV/cm]

  // Make a sensor
  auto sensor = new Sensor();
  sensor->AddComponent(comp);

  // Track class
  auto track = new TrackHeed();
  track->SetSensor(sensor);
  track->SetParticle(pid.data());
  track->SetMomentum(p);

  // Output file
  std::stringstream foutstr;
  foutstr << "edep" << "."      << pid.data() 
	            << ".p"     << std::setw(4) << std::setprecision(0)
                                << std::setiosflags(std::ios::fixed) 
				<< std::setfill('0')<< pGeV
                    << ".theta" << std::setw(4) << std::setprecision(2)
                                << std::setiosflags(std::ios::fixed) 
				<< std::setfill('0') << thetadeg
                    << ".phi"   << std::setw(5) << std::setprecision(1)
                                << std::setiosflags(std::ios::fixed) 
				<< std::setfill('0') << phideg
                    << ".dlen"  << std::setw(6) << std::setprecision(2) 
                                << std::setiosflags(std::ios::fixed)
			       	<< std::setfill('0') << dlen
		    << ".root" << std::ends;
  auto fp   = new TFile(foutstr.str().data(),"RECREATE","Marilin");
  auto tp   = new TTree("Event","Event");
  auto evp  = new GEN::Event;
  tp->Branch("Event","GEN::Event",&evp,64000,2);

  // fill tree
  track->EnableDebugging();
  for (int i = 0; i < nEvents; ++i) {
    if (i == 1) track->DisableDebugging();
    if (i % 1000 == 0) std::cout << i << "/" << nEvents << std::endl;
    // Initial position and direction 
    double x0 = 0., y0 = 0., z0 = dlen, t0 = 0.; // track goes through origin
    double dx0 = TMath::Cos(theta)*TMath::Sin(phi);
    double dy0 = TMath::Cos(theta)*TMath::Cos(phi);
    double dz0 = TMath::Sin(theta);
    double s0 = (-(width/2)-y0)/dy0; // track starts from y = -width/2 at t = t0
           x0 += dx0*s0;
           y0 += dy0*s0;
           z0 += dz0*s0;
    double p0 = track->GetMomentum();

    track->NewTrack(x0, y0, z0, t0, dx0, dy0, dz0);
    GEN::Track tpctrack(t0, TVector3(x0,y0,z0), TVector3(p0*dx0, p0*dy0, p0*dz0), pid, width);

#if 0
    std::cerr << "break 1" << std::endl;
#endif
    // Cluster coordinates
    double xc = 0., yc = 0., zc = 0., tc = 0.;
    // Number of electrons produced in a collision
    int nc = 0;
    // Energy loss in a collision
    double ec = 0.;
    // Dummy variable (not used at present)
    double extra = 0.;

    GEN::Hit hit(&tpc);

#if 0
    std::cerr << "break 2" << std::endl;
#endif
    // Loop over the clusters.
    while (track->GetCluster(xc, yc, zc, tc, nc, ec, extra)) {
      ec *= 1.e-3; // [eV] to [keV]
      auto cp = new GEN::Cluster(tc, TVector3(xc, yc, zc), &tpctrack, ec);       
      for (int ie = 0; ie < nc; ie++) {
        double xe, ye, ze, te;
        double ee, dxe, dye, dze;
        track->GetElectron(ie, xe, ye, ze, te, ee, dxe, dye, dze);  
#if 0
        auto ep = new GEN::Electron(te, TVector3(xe, ye, ze), TVector3(dxe, dye, dze), ee, cp);
#else // Include magnetic field effect
        auto ep = new GEN::Electron(te, TVector3(xc, yc, ze), TVector3(dxe, dye, dze), ee, cp);
#endif
        cp->Add(ep);
      }
      tpctrack.Add(cp); 
    }
#if 0
    std::cerr << "#clusters = " << tpctrack.GetEntries() << std::endl;
#endif
#if 0
    std::cerr << "break 3" << std::endl;
#endif
    tpc.Drift(tpctrack); //Drift track
    TIter nextcl(&tpctrack);
    while (auto cp = static_cast<GEN::Cluster *>(nextcl())) {
      auto acp = new GEN::Cluster(cp,&hit);
      if (acp->GetEntries()) hit.Add(acp);
      else                   delete acp;
    }
#if 0
    std::cerr << " N = "    << hit.GetNsum()   << " Nc = " << hit.GetEntries() 
              << " Q = "    << hit.GetCharge() 
	      << " xcog = " << hit.GetXcog()   << " tcog = " << hit.GetTcog() 
              << std::endl;
#endif
    evp->SetEventNo(i);
    evp->SetTrackPtr(&tpctrack);
    evp->SetHitPtr(&hit);

    tp->Fill();
  }
  std::cout << nEvents << "/" << nEvents << std::endl;
#if 0
    std::cerr << "break 4" << std::endl;
#endif
  tp->Write();
  fp->Write();
  fp->Close();

  std::cerr << "That's it, folks" << std::endl;

  //app.Run(kTRUE); 
}
