#include <vector>
#include <algorithm>
#include <iostream>
#include <sstream>
#include <fstream>

#include <TRandom.h>
#include <TFile.h>
#include <TTree.h>
#include <TNtupleD.h>

#include "Hit.h"
#include "Electron.h"
#include "Cluster.h"
#include "Track.h"
#include "Event.h"
#include "MPGD.h"

using namespace std;

void neff(const char *fname   = "Data/edep.electron.p0005.theta0.00.phi000.0.dlen025.0.root", 
                int   maxnevt = 999999999)
{
  // ---------------
  // Open input file
  // ---------------
  cerr << "Infile: " << fname << endl;
  TFile &file = *new TFile(fname);
  cerr << "File opened" << endl;

  // ---------------
  // Set up TTree
  // ---------------
  TTree *tp = static_cast<TTree *>(file.Get("Event"));
  GEN::Event   *evtp = nullptr;
  tp->SetBranchAddress("Event", &evtp);
  int nlist = tp->GetEntries();
  //cerr << "nlist = " << nlist << endl;
  maxnevt = TMath::Min(maxnevt, nlist);
 
  //--
  // Retrieve job parameters from the input file.
  //--
  tp->GetEntry(0);
  GEN::Track *trp   = evtp->GetTrackPtr();
  TVector3    pv    = trp->GetP0();
  GEN::Hit   *htp   = evtp->GetHitPtr();

  const GEN::MPGD *mpgdp       = htp->GetMPGDPtr();
  const GEN::MPGD &mpgd        = *mpgdp;
        double     padrowpitch = mpgd.GetPadRowPitch();
        double     polyatheta  = mpgd.GetTheta();

  const double kPi   = acos(-1.);
  std::string pid    = trp->GetParticleName(); // particle name
  double      dlen   = trp->GetX0().Z();       // drift length [cm]
  double      p      = pv.Mag()/1.e9;          // momentum [eV]

  double      phi    = atan2(pv.X(),pv.Y());   // phi
  double      theta  = atan2(pv.Z(),pv.Y());  // theta
              phi   *= 180./kPi;
              theta *= 180./kPi;

  std::cerr << "========================================================" << std::endl;
  std::cerr << " nEvents     = " << maxnevt                 << std::endl
            << " dlen        = " << dlen        << " [cm]"  << std::endl
            << " phi         = " << phi         << " [deg]" << std::endl
            << " theta       = " << theta       << " [deg]" << std::endl
            << " p           = " << p           << " [GeV]" << std::endl
            << " pid         = " << pid                     << std::endl
            << " padrowpitch = " << padrowpitch << " [cm]"  << std::endl
            << " polyatheta  = " << polyatheta              << std::endl;
  std::cerr << "========================================================" << std::endl;

  // ----------------
  // Open output file
  // ----------------
  std::stringstream foutstr;
  foutstr << "tup"  << "."      << pid.data() 
	            << ".p"     << std::setw(4) << std::setprecision(0)
                                << std::setiosflags(std::ios::fixed) 
				<< std::setfill('0')<< p
                    << ".theta" << std::setw(4) << std::setprecision(2)
                                << std::setiosflags(std::ios::fixed) 
				<< std::setfill('0') << theta
                    << ".phi"   << std::setw(5) << std::setprecision(1)
                                << setiosflags(std::ios::fixed) 
				<< std::setfill('0') << phi
                    << ".dlen"  << std::setw(6) << std::setprecision(2) 
                                << setiosflags(std::ios::fixed)
			       	<< std::setfill('0') << dlen
		    << ".root" << std::ends;
  TFile &ofile = *new TFile(foutstr.str().data(),"RECREATE","Marilin");

  //--
  // Define n-tuple
  //--
  stringstream tuplist;
  tuplist << "nev:dlen:phi:theta:p"<< ":"
	  << "padrowpitch:xcog:tcog:nsum:k2sum" << ":" 
          << "qsum:q2sum:qc2sum"
	  << ends;
  TNtupleD tup("tup","",tuplist.str().data()); 

  // ---------------
  // Loop over hits
  // ---------------
  int nevts = 0;
  for (int i=0; i<maxnevt; i++) {
    tp->GetEntry(i);
    GEN::Hit *hitp = evtp->GetHitPtr();
    double qsum = hitp->GetCharge();
    double nsum = hitp->GetNsum();
    double xcog = hitp->GetXcog();
    double tcog = hitp->GetTcog();
#if 0
    cerr << " xcog = " << xcog
         << " tcog = " << tcog
         << " nsum = " << nsum               // # signal e's in this hit
	 << " qsum = " << qsum               // charge sum
	 << " nc = "   << hitp->GetEntries() // # clusters in this hit
	 << endl;
#endif
    double k2sum   = 0;
    double qc2sum  = 0;
    double q2sum   = 0;
    TIter next(hitp);
    // Loop over clusters
    while (GEN::Cluster *cp = static_cast<GEN::Cluster *>(next())) {
      int    kc = cp->GetEntries();      // # signal e's from this cluster
      double yc = cp->GetPosition().Y(); // y position of this cluster
      k2sum += kc*kc;
#if 0
      cerr << "   kc = " << kc 
           << "   yc = " << yc
           << endl;
#endif
      double qc    = 0;
      TIter nexte(cp);
      while (GEN::Electron *ep = static_cast<GEN::Electron *>(nexte())) {
        double q = ep->GetAmpFactor();
	q2sum += q*q;
	qc    += q;
      }
      qc2sum += qc*qc;
    }
    //--
    // Fill n-tuple
    //--
    nevts++;
    const int kNdata = 100;
    double data[kNdata];
    data[ 0] = nevts;
    data[ 1] = dlen;
    data[ 2] = phi;
    data[ 3] = theta;
    data[ 4] = p;
    data[ 5] = padrowpitch;
    data[ 6] = xcog;
    data[ 7] = tcog;
    data[ 8] = nsum; 
    data[ 9] = k2sum;
    data[10] = qsum;
    data[11] = q2sum;
    data[12] = qc2sum;
    tup.Fill(data);
  }
  // ---------------
  // Write results
  // ---------------
  tup.Write();
#if 0
  tup.Draw("nsum>>hNsum(500,0.500)","","goff");
  auto   hNsum  = static_cast<TH1D *>(gROOT->FindObject("hNsum"));
  double Nav    = hNsum->GetMean();

  tup.Draw("1./nsum>>hNinv(500,0.500)","","goff");
  auto   hNinv  = static_cast<TH1D *>(gROOT->FindObject("hNinv"));
  double Neff   = 1./hNinv->GetMean();

  tup.Draw("q2sum>>hQ2sum(500,0.5e9)","","goff");
  auto   hQ2sum = static_cast<TH1D *>(gROOT->FindObject("hQ2sum"));
  double q2tot  = hQ2sum->GetMean()*hQ2sum->Integral();
  tup.Draw("qsum>>hQsum(500,0.5e5)","","goff");
  auto   hQsum  = static_cast<TH1D *>(gROOT->FindObject("hQsum"));
  double qtot   = hQsum->GetMean()*hQsum->Integral();
  double ntot   = Nav*hNsum->Integral();
  double Neffg  = Neff/(ntot*q2tot/(qtot*qtot));

  tup.Draw("k2sum/nsum/nsum>>hK2sum(500,0.500)","","goff");
  auto   hK2sum = static_cast<TH1D *>(gROOT->FindObject("hK2sum"));
  double Nheff  = 1./hK2sum->GetMean();
  
  tup.Draw("qc2sum/qsum/qsum>>hFqc2(100,0.,1.)","","goff");
  auto   hFqc2  = static_cast<TH1D *>(gROOT->FindObject("hFqc2"));
  double Nheffg = 1./hFqc2->GetMean();

  tup.Draw("xcog>>hXcog(500,-0.1,0.1)","","goff");
  auto   hXcog  = static_cast<TH1D *>(gROOT->FindObject("hXcog"));
  double sgx    = hXcog->GetRMS();
  double dsgx   = hXcog->GetRMSError();
  if (sgx > 0.) {
    //hXcog->Fit("gaus","Q","goff");
    hXcog->Fit("gaus","","goff");
    sgx    = hXcog->GetFunction("gaus")->GetParameter(2);
    dsgx   = hXcog->GetFunction("gaus")->GetParError (2);
  }

  cerr << " Nav = " << Nav << " Neff = " << Neff << " Neffg = " << Neffg
       << " Nheff = " << Nheff << " Nheffg = " << Nheffg 
       << " sgx = " << sgx << " dsgx = " << dsgx
       << endl;
#endif
  // ---------------------
  // Close the input file
  // ---------------------
  file.Close();
  //--
  // That's it folks!
  //--
  cerr << "File closed. That's it." << endl;
}
